﻿#requires -version 2
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#                                    pam_esp_call.ps1
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#
#  This script can be used to launch PAM jobs from ESP with advanced logging of error codes
#  
#  The script calls CA ESP Agent's ITPAM Java wrapper, passing it the trace switch. The trace switch
#  enables output of the SOAP request/response stucture. The output is then redirected to a file.
#
#  The trace file is then analyzed. We try to extract 2 PAM variables: ESPExitStatusCode & ESPExitReason
#  If these variables are found:
#         ESPExitStatusCode is used as an exit code to pass back to ESP
#         ESPExitReason is used as a message printed to STDOUT (and therefore can be seen in the ESP spool file)
#
########################################################################################################
#
#  Mapping Config File:
#     A config file can be used to map ESP jobnames to PAM Proces Paths. This can be used as a convenience 
#     when defining ESP jobs and allows for PAM Process paths to be changed without affecting ESP jobs. 
#     
#     Ex Config:
#       JobName,JobPath
#       test,/CN/PathTo/TestProcess
#     
#     This entry means ESP can call:
#         pam_esp_call.ps1 -job test
#
#     which will be translated to run the PAM Process /CN/PathTo/TestProcess
#
########################################################################################################
#
#  Secure Config File:
#     In order to "hide" the esp integration account passwords from this script, we leverage powershell's
#     secure strings. If you run this script without specifiying a username/password, the esp schedule 
#     account will be used. Secure strings are loaded from the config file and decrypted using the
#     logged on user's encryption keys ( an encryption key is unique per user & per server ).
#
#     Ex Config:
#       Hostname,Environment,Username,SecureString
#       mtl-hq-a991,prd,999999,securestingxyz
#
#     This entry will decrypt 'securestingxyz' using the decryption key for user '999999' on server 'mtl-hq-a991'.
#     The resulting password is used to connect to PAM using the 'prd' esp schedule account.
#  
#########################################################################################################
#
#  Requirements: 
#     CA Workload Automation Agent must be installed with the ITPAM plugin wrapper
#
########################################################################################################
#
#  Author: Ian Rich
#  Date:   20130910
#
#  Modifications:
#   20131008 - James Dafoe: 1 - Code refactoring
#                           2 - Use XML object to parse trace log file
#                           3 - Read secure strings from config file and convert to password
#                           4 - Read ESP to PAM job mapping from mapping config file
#                           5 - Ensure Powershell positional parameters are not used
#                           6 - Find path to ESP install location from service definition
#
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Script Parameters (NOTE: must be first executable line)
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
param(
  $dummy,                 # Force non positional params
  [string]$environment,   # Environment
  [string]$user,          # User
  [string]$pass,          # Pass
  [string]$jobname,       # PAM Job
  [string]$params         # PAM Job Params
)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Script Behaviour
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Set-StrictMode -version 2.0

$DebugPreference       = "SilentlyContinue"
$ErrorActionPreference = "stop"


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Global Variables
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

# Script Invocation
$SCRIPT_NAME = $MyInvocation.MyCommand.Name
$SCRIPT_PATH = Split-Path -parent $MyInvocation.MyCommand.Definition

# Random logfile to store output from trace mode
$LOGFILE = "D:\CA\Scripts\Temp\"+ "{0:D4}" -f (1..9999 | Get-Random) + (Get-Date).Ticks + ".log"

# CSV Conf Files
$SECURE_CONF = $SCRIPT_PATH + '\secure.conf'
$MAPPING_CONF = $SCRIPT_PATH + '\job_mapping.conf'


# DEV/PRD Orchestrator Definitions
$ORCHESTRATORS = @{ 'DEV' = @{  'USER' = 'sch-esp-capad';
                                'ENDPOINT' = 'https://automation-dev.domain.com:443/itpam/soap';
                                'SERVERS' = @( 'mtl-hq-d991', 'mtl-hq-d992' )
                             };
                    'PRD' = @{  'USER' = 'sch-esp-capa';
                                'ENDPOINT' = 'https://automation.domain.com:443/itpam/soap';
                                'SERVERS' = @( 'mtl-hq-a991', 'mtl-hq-a992' )
                             };
}


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#/*-  FUNCTIONS
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++



#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#/*-                                              deleteFile
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#
#  Description : Deletes a file if it exists
#
#===============================================================================================================
function deleteFile( $file ) {
  If ( Test-Path $file ) {
    Remove-Item $file
  }
}



#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#/*-                                              printUsage
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#
#  Description : Prints Script Usage
#
#===============================================================================================================
function printUsage() {
    write-host ""
    write-host "Script Usage:"
    write-host "$SCRIPT_NAME -job JOB [-env ENV] [-user USER] [-pass PASS] [-params <name=value::name2=value2>]"
    write-host ""
    write-host "  -job         '<JOB>'                        Path to PAM Process or Job from mapping table"
    write-host "  -env         '<ENV>'                        (Optional on Orchestrator) Environment DEV|PRD"
    write-host "  -user        '<USER>'                       (Optional on Orchestrator) PAM Username"
    write-host "  -pass        '<PASSWORD>'                   (Optional on Orchestrator) PAM Password"
    write-host "  -params      '<name=value::name2=value2>'   (Optional) Parameters to be passed to PAM job"   
    write-host "                                              Parameter format:"
    write-host "                                              name1=value1::name2=value2"
    write-host "                                              Where :: is used to separate multiple parameters"
}



#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#/*-                                              callPAMWrapper
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#
#  Description : Calls the ESP PAM Java Wrapper
#
#  Parameters : $logfile:     Logfile to copy output from call
#               $endpoint:    PAM SOAP URL to connect to
#               $user:        PAM Username
#               $pass:        PAM Password
#               $job:         PAM process path to call
#               $paramsArray: Parameters to pass to PAM. Array Format ->  @( 'param[name]=value', 'param[name]=value', ... )
#
#
#===============================================================================================================
function callPAMWrapper( $logfile, $endpoint, $user, $pass, $job, $paramsArray ) {

  ##################################################################
  # Build command to call ESP PAM Wrapper, then call it
  ################################################################## 
  
  # Find ESP install location from service
  try {
    # ESP Account can't query WMI ( Permissions Error ), so hardcoding path
    #$espIntsall = ((Get-WmiObject -Class Win32_Service -Filter "Name like 'CA Workload Automation Agent%'").PathName -Split " ")[0] -replace '"','' | 
    #                  Split-Path -Parent
    $espIntsall = 'D:\ESP11.3'
  }
  catch {
    Write-Host 'CA Workload Automation Agent must be installed with itpam wrapper in order to run this script ... '
    $host.SetShouldExit( 255 )
    exit
  }
  
  # Set java and java class path
  # Class Path: -cp D:\ESP11.3\jars\library.jar;%LIBS%\pluginlib.jar;D:\ESP11.3\jars\ext\jsafeJCEFIPS.jar;D:\ESP11.3\jars\commonsvc_log.jar;D:\ESP11.3\jars\itpam.jar',
  # Class: cybermation.plugins.itpam.driver.ITPAMDriver'
  $java =  "$espIntsall\jre\bin\java.exe";
  
  $args = @( '-cp',
             "$espIntsall\jars\*;$espIntsall\jars\ext\*;%LIBS%\pluginlib.jar",    # Java Class path
             'cybermation.plugins.itpam.driver.ITPAMDriver'                       # Java Class to call
             "--user=$user",                                                      # User
             "--password=$pass",                                                  # Password
             "--endpointaddress=$endpoint",                                       # SOAP URL
             "--operation=exec_process",                                          # Start Process Switch
             "--mainarg[name]=$job",                                              # Path to PAM Process
             "--operation_switch[track]=true",                                    # Track in ESP until job completes
             "--trace")                                                           # Verbose trace mode enabled  
  
  # Run it!
  &$java ($args + $paramsArray) | Out-File $logfile  
  return $LASTEXITCODE
}



#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#/*-                                            analyzeLog
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#
#  Description : Parses the log file generated by call to PAM wrapper ( with --trace switch )
#                Tries to find ESPExitStatusCode & ESPExitReason variables from PAM to pass back to ESP
#
#  Parameters :  $logfile:    The logfile to analyze
#                $exitCode:   The exit code from call to PAM wrapper that generated logfile
#
#===============================================================================================================
function analyzeLog( $logFile, $exitCode ) {

  ##################################################################
  # Wrapper and PAM Process finished successfully
  ################################################################## 
  if ( $exitCode -eq '0' ) {
    $espExit = $exitCode  
    $espMsg = 'PAM Process completed successfully'
  } 
  
  ##################################################################    
  # Wrapper reported exit code
  # Analyze Log File
  ################################################################## 
  else { 
    # Read the log file
    ################################################################## 
    $logContents = Get-Content $logFile
    
    # Check for error submitting request to PAM ( ex: Unable to find process path )
    ################################################################## 
    if ( $exitCode -eq '3' ) {  
      $espMsg = $logContents | 
                  Where-Object { $_.Contains('Error submitting request:') } | 
                    Select-Object -Last 1
                    
      if ( $espMsg -eq $null ) {
        $espMsg = ''
        Write-Host 'Unable to parse the error from the log file ... '
      }
      
      $espExit = $exitCode
    }
      
    # Check for error with call to Wrapper
    # Analyze first line of logfile
    # ( NOTE: This should never happen because we validate parameters before calling Wrapper
    # but doesn't hurt to check it here ) 
    ################################################################## 
    elseif ( $logContents[0] -Match '(\*\*\* Unrecognized argument:.*|\*\*\* Invalid argument\. Must start with -- \*\*\*)' ) {
      $espExit = $exitCode
      $espMsg = ''
      Write-Host "Failed calling PAM Wrapper '$Matches[1]'. Please validate your parameters: '$params'"  
    }
    
    # Wrapper called succesfully
    # Wrapper called PAM succesfully
    # BUT PAM process failed
    # Try to parse out ESPExitStatusCode & ESPExitReason
    ################################################################## 
    else {
      # Start with same exit code by default
      $espExit = $exitCode
      $espMsg = ''
  
      # Try to find last XML reponse and convert it to XML
      $isXMLValid = $false
      try {
        $xml = $logContents -join "`n" | Select-String '(?smi)<\?xml.*?</SOAP-ENV:Envelope>' -AllMatches
        $xml = [xml] $xml.Matches[-1].Value
        $isXMLValid = $true
      }
      catch {
        # Invalid XML
        write-host 'Unable to parse XML log fle ... '
        $isXMLValid = $false
      }
    
      if ( $isXMLValid -eq $true ) {
        # Parse ESPExitStatusCode
        ################################################################## 
        $espExit = ( $xml.Envelope.Body.CheckFlowResponse.CheckFlowResult.params.param | 
                        Where-Object {$_.name -eq 'ESPExitStatusCode'} ) 
        
        if ( $espExit -eq $null ) {
          $espExit = $exitCode 
          write-host 'Unable to parse ESPExitStatusCode ...'    
        }
        else {
          $espExit = $espExit.get_innerXml()
        
          if ( -not( $espExit -match '\d+') ) {
            $espExit = $exitCode
            write-host 'Invalid ESPExitStatusCode (not an integer) ...'
          }
        }
        
        # Parse ESPExitReason
        ################################################################## 
        $espMsg = ( $xml.Envelope.Body.CheckFlowResponse.CheckFlowResult.params.param | 
                      Where-Object {$_.name -eq 'ESPExitReason'} )
                          
        if ( $espMsg -eq $null ) {
          $espMsg = ''
          write-host 'Unable to parse ESPExitReason ...'
        }
        else {
          Add-Type -AssemblyName System.Web
          $espMsg = [System.Web.HttpUtility]::HtmlDecode( $espMsg.get_innerXml() )
        }
      }
    }
  } 
  
  # Set exit code for ESP
  ##################################################################
  Write-Host "`n"
  Write-Host 'ESP Error Code:'
  Write-Host '============================='
  Write-Host 'Wrapper Reported: ' $exitCode
  Write-Host 'Sending to ESP:   ' $espExit
  Write-Host "`n"
  Write-Host 'ESP Message:'
  Write-Host '============================='
  Write-Host $espMsg
  $host.SetShouldExit( $espExit )
}





#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#/*- MAIN
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function main( $environment, $user, $pass, $jobname, $paramsString ) {

  # Get computername/username we're running on
  $currentHostname = $env:computername
  $currentUsername = $env:username
 
  ##################################################################
  # Parse & Validate Command Line
  ##################################################################   
  
  #---------------
  # Environment
  #--------------- 
  # User did NOT supply environment ...
  # Check if we're running on an Orchestrator node. If we are, set the environment
  if ( $environment -eq '' ) {
    if ( $ORCHESTRATORS.DEV.SERVERS -Contains $currentHostname ) {
      $environment = 'DEV'
    }
    elseif ( $ORCHESTRATORS.PRD.SERVERS -Contains $currentHostname ) {
      $environment= 'PRD'
    }
    else {
      Write-Host "You are not running on an Orchestrator node. -env must be specified ..."
      printUsage
      $host.SetShouldExit( 255 )
      exit 
    }
  }
  # User supplied an environment. Validate it
  else {
    if ( -not $ORCHESTRATORS.Contains( $environment ) ) {
      Write-Host 'Invalid environment specified (-env) ...'
      printUsage  
      $host.SetShouldExit( 255 )
      exit      
    }
  }
  
  # Set Orchestrator SOAP URL to connect to
  $endPoint = $ORCHESTRATORS[$environment].ENDPOINT
  

  #---------------
  # User & Pass
  #---------------
  # If user or pass supplied, they must be supplied together
  if ( ($user -ne '' -and $pass -eq '') -or
       ($pass -ne '' -and $user -eq '') ) {
    Write-Host '-user & -pass must be supplied together ...'
    printUsage
    $host.SetShouldExit( 255 )
    exit
  }
  # User & Pass NOT supplied by user
  elseif ( $user -eq '' -and $pass -eq '' ) {
    # Get default user
    $user = $ORCHESTRATORS[$environment].USER
    
    # Get default password ( decrypt securestring from config file )
    if ( Test-Path -Path $SECURE_CONF -PathType leaf ) {
      try {
        $secureString = @( Import-Csv -Path $SECURE_CONF | 
                              Where-object {  ($_.Hostname -eq $currentHostname) -and 
                                              ($_.Username -eq $currentUsername) -and
                                              ($_.Environment -eq $environment) } )
                                  
      }
      catch {
        write-host "Secure string config file not accessible or not formatted correctly ..."
        $host.SetShouldExit( 255 )
        exit
      }
      
      if ( -not $secureString ) {  
        write-host "Unable to find entry in secure string config file for '$currentUsername' on '$currentHostname' for '$environment' ..."
        $host.SetShouldExit( 255 )
        exit
      }
      elseif ( $secureString.Length -ne 1 ) {
        write-host "Multiple entries found in secure string config file for '$currentUsername' on '$currentHostname' for '$environment' ..."
        $host.SetShouldExit( 255 )
        exit    
      }
      else {
        try {
          $secure = $secureString[0].SecureString | ConvertTo-SecureString
        }
        catch {
          Write-Host "Invalid secure string found in secure string config file for '$currentUsername' on '$currentHostname' for '$environment' ..."
          $host.SetShouldExit( 255 )
          exit   
        }
        
        $pass = (New-Object System.Management.Automation.PSCredential -ArgumentList $user, $secure).GetNetworkCredential().Password
      }
    }
    else {
      Write-Host 'Unable to find secure string configuration file ...'
      $host.SetShouldExit( 255 )
      exit
    } 
  }
   

  #---------------
  # Job Name
  #---------------
  if ( $jobname -eq '' ) {
    Write-Host "-job is a required parameter" ""
    printUsage
    $host.SetShouldExit( 255 )
    exit
  }
  else {
    # Try to look up the job in the ESP to PAM job mapping table
    if ( Test-Path -Path $MAPPING_CONF -PathType leaf ) {
      try {
        $jobPath = @( Import-Csv -Path $MAPPING_CONF | 
                        Where-object { ($_.JobName -eq $jobname) } )
                        
        if ( -not $jobPath ) {  
          write-host 'Unable to find jobname in ESP to PAM job mapping config file. Using jobname passed to script ...'
          $jobPath = $jobname
        }
        elseif ( $jobPath.Length -ne 1 ) {
          write-host 'Multiple job entries found in ESP to PAM job mapping config file. Using jobname passed to script ...'
          $jobPath = $jobname        
        }
        else {  
          $jobPath = $jobPath[0].JobPath 
        }                            
      }
      catch {
         write-host 'ESP to PAM job mapping config file not accessible or not formatted correctly. Using jobname passed to script ...'
         $jobPath = $jobname    
      }
    }
    else {
      # No configfile found, use job name that was passed in
      Write-Host 'No ESP to PAM job mapping conf file Found. Using jobname passed to script ...'
      $jobPath = $jobname
    }
  }

  
  #---------------
  # Optional Params
  #---------------  
  $paramsArray = @()
  if ( $paramsString -ne '' ) {
    $arguments = $params -Split '::'
    
    foreach( $argument in $arguments ) {
      # Split each ESP parameter into the variable name, and the variable value and change it to the correct format
      # Format to send to Java wrapper is: --param[Server2Backup]=mtl-hq-f20
      try { 
        $varName = $argument.Substring( 0, $argument.IndexOf('=') )
        $varValue = $argument.Substring(   ($argument.IndexOf('=') + 1), 
                                           ($argument.Length - $argument.IndexOf('=') - 1)   )
                                         
        if ( $varName -eq '' -or $varName -eq $null -or 
              $varValue -eq '' -or $varValue -eq $null ) {
              
          throw 'Parse Parameters Error'
        }
        $paramsArray += '--param[' + $varName + ']=' + $varValue
      }
      catch {
        write-host "Unable to parse parameters: '$paramsString'"
        printUsage
        $host.SetShouldExit( 255 )
        exit
      }
    } 
  }    
    
  ##################################################################
  # 1 - Call ESP PAM Wrapper, writing output to log
  # 2 - Analyze log 
  # 3 - Delete log
  ################################################################## 
  $exitCode = callPAMWrapper $LOGFILE $endPoint $user $pass $jobPath $paramsArray
  analyzeLog $LOGFILE $exitCode
  deleteFile $LOGFILE
}  


# Main Entry Point
if ( $dummy ) {
  Write-Host 'Error: You can not use positional parameters'
  printUsage
  $host.SetShouldExit( 255 )
  exit
}
main $environment $user $pass $jobname $params